//
// Avi2Cvc D by ssjx (http://ssjx.co.uk)
//
// Compile with:	
//		dmd -i avi2cvc.d
//
// Version 2.9 C# (25/05/23)
//
// Version 2.8 D (25/02/23)
// 	25/02/2023 - Source clean up, now Avi2Cvc/Bmp2Cvi share a lot of functions
//
// Version 2.7 D (16/02/23)
//
//
//
using System;
using System.IO;

class Avi2cvc
{
	static byte[] gray=new byte[16000];
	static byte[] pic=new byte[4000];

	static int swp(byte a,byte b,byte c,byte d){
		int ta = a;
		int tb = b;
		int tc = c;
		int td = d;
		return ta + (tb << 8)+ (tc << 16)+ (td << 24);
	}

	static int byte_swap(byte a,byte b){
		int ta = a;
		int tb = b;
		return ta + (tb << 8);
	}

	static void Main(string[] args)
	{
		int dither=1;
		string fn="";
		
		Console.WriteLine("Avi2Cvc v2.9 C# (22/05/23) by ssjx (http://ssjx.co.uk)");
		Console.WriteLine("------------------------------------------------------");
		
		if (args.Length>0){
			foreach(var par in args){
				if (par[0]=='-'){
					if (par=="-nodither"){dither=0;}
					if (par=="-ordered"){dither=1;}
					if (par=="-diffuse"){dither=2;}
				}else{
					// Not the exe or an option so probably an avi file
					fn=par;
				}
			}
		}else{
			Console.WriteLine("Usage: avi2cvc [option] myfile.avi");
			Console.WriteLine("Options: -nodither , -ordered (default), -diffuse");
			Console.WriteLine("Avi should be 160x100 and either uncompressed or MSVC compressed.");
			return;
		}

		// Start reading AVI file
		if (File.Exists(fn)){	
			string outfn=fn+".cvc";	
		
			Console.WriteLine("In: ",fn);
			Console.WriteLine("Out: ",outfn);
			Console.WriteLine("Dither: ",dither);
		
			try{
				int bmhd_w=0;
				int bmhd_h=0;
				int bmhd_d=0;
				
				int bmhd_frames=0;
				int bmhd_fps=0;
	
				//
				// New AVI parser
				// 
				
				Stream infile = File.OpenRead(fn);
					
					bool ismovie=false;	
					string ctype;
					string handler="";
					
					byte[] avi_list=new byte[4];
					int size=0;
					
					infile.Seek(12,SeekOrigin.Current);
					
					string subtype="";
					do{
						infile.Read(avi_list,0,avi_list.Length);
						ctype=""+(char)avi_list[0]+(char)avi_list[1]+(char)avi_list[2]+(char)avi_list[3];
						Console.WriteLine(ctype);
							
						if (ctype.Equals("LIST")){
							byte[] avi_chunk=new byte[8];
							infile.Read(avi_chunk,0,avi_chunk.Length);
						
							size=swp(avi_chunk[0],avi_chunk[1],avi_chunk[2],avi_chunk[3]);
							ctype=""+(char)avi_chunk[4]+(char)avi_chunk[5]+(char)avi_chunk[6]+(char)avi_chunk[7];
							Console.WriteLine("List : "+ctype);
						
							switch (ctype){
								case "movi":
									Console.WriteLine("Movie List!");
									ismovie=true;
								break;
								case "hdrl":
									Console.WriteLine("Header List");
								break;
								case "strl":
									Console.WriteLine("Stream List");
								break;
								default:
									Console.WriteLine("Unknown List "+ctype);
								break;
							}
						}else{
							// Chunk
							byte[] avi_chksize=new byte[4];
							infile.Read(avi_chksize,0,avi_chksize.Length);
							size=swp(avi_chksize[0],avi_chksize[1],avi_chksize[2],avi_chksize[3]);
							
							switch (ctype){
								case "avih":
									// AVI file header
									Console.WriteLine("+ Avi Header");
									byte[] avi_header=new byte[56];
									infile.Read(avi_header,0,avi_header.Length);
							
									bmhd_frames=swp(avi_header[16],avi_header[17],avi_header[18],avi_header[19]);
									Console.WriteLine("  Frames: "+bmhd_frames);
								break;
								case "strh":
									// Stream header
									Console.WriteLine("+ Stream Header");
									byte[] avi_strh=new byte[56];
									infile.Read(avi_strh,0,avi_strh.Length);
									
									subtype=""+(char)avi_strh[0]+(char)avi_strh[1]+(char)avi_strh[2]+(char)avi_strh[3];
									
									 if (subtype.Equals("vids")){
										handler=""+(char)avi_strh[4]+(char)avi_strh[5]+(char)avi_strh[6]+(char)avi_strh[7];
										//bmhd_comp=handler;
										handler=handler.ToLower();
										Console.WriteLine("  Format: ["+handler+"]");	
										
										int scale=swp(avi_strh[20],avi_strh[21],avi_strh[22],avi_strh[23]);
										int fps=swp(avi_strh[24],avi_strh[25],avi_strh[26],avi_strh[27]);
										
										bmhd_fps=fps/scale;
										
										Console.WriteLine("  Rate: "+fps+" / "+scale+" ["+bmhd_fps+"]");
										
									}else{
										Console.WriteLine("  Ignoring: "+subtype);
									}
									
									// Console.WriteLine("Rate: ",vheader[0].fps," / ",vheader[0].scale);
								break;
								case "strf": 
									// Stream format
									Console.WriteLine("+ Stream Format");
									byte[] avi_strf=new byte[size];
									infile.Read(avi_strf,0,avi_strf.Length);
									
									if (subtype.Equals("vids")){
										bmhd_w=swp(avi_strf[4],avi_strf[5],avi_strf[6],avi_strf[7]);
										bmhd_h=swp(avi_strf[8],avi_strf[9],avi_strf[10],avi_strf[11]);
										bmhd_d=byte_swap(avi_strf[14],avi_strf[15]);
										Console.WriteLine("  Size: {0} x {1} x {2}",bmhd_w,bmhd_h,bmhd_d);
									}else{
										Console.WriteLine("  Ignoring: {0}",subtype);
									}
								break;
								case "JUNK":
									Console.WriteLine("+ Junk");
									infile.Seek(size,SeekOrigin.Current);
								break;
								default:
									Console.WriteLine("+ Unknown {0} Size: {1}",ctype,size);
									infile.Seek(size,SeekOrigin.Current);
								break;
							}
						}
						
						if (size==0){
							Console.WriteLine("Panic Exit!");
							ismovie=true;
						}
							
					}while(ismovie==false);	
								
				//
				if ((handler!= "dib ") && (handler!="msvc")) {
					Console.WriteLine("Unsupported compression [", handler, "], video must be uncompressed DIB or MSVC.");
					return;
				}
			
				int compression=-1;
				if (handler.Equals("dib ")==true && bmhd_d==24){compression=0;}
				if (handler.Equals("msvc")==true && bmhd_d==16){compression=1;}
			
				// Make sure AVI is correct resolution
				if ((bmhd_w > 160) || ( bmhd_h > 100)) {
					Console.WriteLine("Resolution must be 160x100 or lower.");
					return;
				}

				if ((bmhd_w < 0) || ( bmhd_h < 0)) {
					Console.WriteLine("Resolution must be 160x100 or lower.");
					return;
				}

				if ((bmhd_w % 4) > 0) {
					Console.WriteLine("AVI width must be divisible by 4");
					return;
				}
				
				//

				if (ismovie==true){
					byte[] cvc_header=new byte[12];
					cvc_header[0]=(byte)'c';
					cvc_header[1]=(byte)'y';
					cvc_header[2]=(byte)'v';
					cvc_header[3]=(byte)'d';	
					
					// Size
					cvc_header[4]=0;
					cvc_header[5]=0;
					cvc_header[6]=0;
					cvc_header[7]=0;
					
					// Lock
					cvc_header[8]=0;
					cvc_header[9]=0;
					
					// Headers (1 = Video) 
					cvc_header[10]=0;
					cvc_header[11]=1;
					
					//
					byte[] cvc_video=new byte[18];
					cvc_video[0]=(byte)'v';
					cvc_video[1]=(byte)'i';
					cvc_video[2]=(byte)'d';
					cvc_video[3]=(byte)'o';
					
					short w=(short)bmhd_w;
					short h=(short)bmhd_h;
					
					// W
					cvc_video[5] = (byte)(w & 0x00ff);
					cvc_video[4] = (byte)((w & 0xff00) >> 8);

					// H
					cvc_video[7] = (byte)(h & 0x00ff);
					cvc_video[6] = (byte)((h & 0xff00) >> 8);
					
					
					//Frames
					cvc_video[11]=(byte)(bmhd_frames &  0x000000ff);
					cvc_video[10]=(byte)((bmhd_frames & 0x0000ff00)>>8);
					cvc_video[9]= (byte)((bmhd_frames & 0x00ff0000)>>16);
					cvc_video[8]= (byte)((bmhd_frames & 0xff000000)>>24);

					// FPS
					cvc_video[12]=(byte)((bmhd_fps & 0xff00)>>8);
					cvc_video[13]=(byte)(bmhd_fps & 0x00ff);
					
					//
					cvc_video[14]=(byte)'c';
					cvc_video[15]=(byte)'y';
					cvc_video[16]=(byte)'v';
					cvc_video[17]=(byte)'1';
					
					byte[] cvc_frame=new byte[6];
					
					try
					{
						byte[] cvc_data;
						
						byte[] gray8=new byte[16000];	// 8bit gray frame
																	
						// File outfile = File(outfn, "wb");  
						Console.WriteLine("Saving as ",outfn,"...");
						Stream outfile = File.Create(outfn);					
						outfile.Write (cvc_header, 0, cvc_header.Length);
						outfile.Write (cvc_video, 0, cvc_video.Length);
						
						//
						int maxsize=(bmhd_w*bmhd_h)/4;
						
						Msvc msvc=new Msvc();
						Cv1 encode=new Cv1();
						Utils utils=new Utils();
					
						// Create empty variable sized previous frame
						encode.prev=new byte[pic.Length];
					
						byte[] frame_chunk=new byte[8];
						byte[] frame_data=new byte[1];	// So it has a size
					
						for(int f=0;f<bmhd_frames;f++){
							// Read a frame from the AVI
							infile.Read(frame_chunk,0,frame_chunk.Length);
							
							size=swp(frame_chunk[4],frame_chunk[5],frame_chunk[6],frame_chunk[7]);
							
							if (size!=frame_data.Length){
								frame_data=new byte[size];
							}
							
							// infile.read(frame_data);
							infile.Read(frame_data,0,frame_data.Length);
							
							switch(compression){
								case 0:
									utils.grey(frame_data,gray,bmhd_w,bmhd_h);
								break;
								
								case 1:
									msvc.decode(frame_data,gray8,bmhd_w,bmhd_h);
									for(int i=0;i<gray8.Length;i++){
										gray[i]=gray8[i];
									}
								break;
							}
							
							switch(dither){
								case 0:
									utils.twobit(gray,bmhd_w,bmhd_h);
								break;
								
								case 1:
									utils.ordered(gray,bmhd_w,bmhd_h);
								break;
								
								case 2:
									utils.diffuse(gray,bmhd_w,bmhd_h);
								break;
								
								default:
									utils.ordered(gray,bmhd_w,bmhd_h);
								break;
							}
							
							// 8bit > 2bit
							utils.frameconv(gray,pic,bmhd_w,bmhd_h);
						
							// See which compression method works the best!
							int esize=encode.encode_rle_size(pic);
							short usecomp=1;	
							int tmpsize=0;
						
							tmpsize=encode.encode_delta1_size(pic);
							if (tmpsize<esize){
								esize=tmpsize;
								usecomp=2;
							}
							
							tmpsize=encode.encode_delta2_size(pic);
							if (tmpsize<esize){
								esize=tmpsize;
								usecomp=3;
							}
						
							// In case compression just does not work (or not worth it)...
							if (esize>=maxsize){
								esize=(short)maxsize;
								usecomp=0;
							}
							
							// Force
							// esize=(short)maxsize;					
							// usecomp=0;
							
							cvc_data=new byte[esize];
							
							switch(usecomp){
							case 0:
								// Uncompressed frame
								encode.encode_none(pic,cvc_data);
							break;
							
							case 1:
								// RLE compressed frame...
								encode.encode_rle(pic,cvc_data);
							break;
							
							case 2:
								// Delta 1 compressed frame...
								encode.encode_delta1(pic,cvc_data);
							break;
							
							case 3:
								// Delta 2 compressed frame...
								encode.encode_delta2(pic,cvc_data);
							break;
							}
						
							// Write  the CVC frame header (3 x short)
							cvc_frame[0]=0;
							cvc_frame[1]=0;
											
							cvc_frame[3] = (byte)(esize & 0x00ff);
							cvc_frame[2] = (byte)((esize & 0xff00) >> 8);
							
							cvc_frame[4] = 0;				//(byte)(usecomp & 0x00ff);
							cvc_frame[5] = (byte)usecomp;	//(byte)((usecomp & 0xff00) >> 8);
					
											
							Console.WriteLine("Frame {0}: {1} - {2} => {3}",f,usecomp,pic.Length,esize);
							
							outfile.Write(cvc_frame, 0, cvc_frame.Length);
							outfile.Write(cvc_data, 0, cvc_data.Length);	
						}
						
					}catch (Exception e){
						Console.WriteLine("Exception caught creating output file!");
						Console.WriteLine(e);
					}
				}
			

			}catch (Exception e){
				Console.WriteLine("Exception caught opening input file!");
				Console.WriteLine(e);
			}
		}else{
			Console.WriteLine("Could not find the file ",fn);
		}
	}
}
