//
// General functions shared by both Avi2Cvc (Video) and Bmp2Cvi (Image) converters
//

using System;	// Abs()

public class Utils{
	
	//
	// Ordered Dithering
	// =================
	// Moved the pattern array out of the function to reduce allocations and make the GC happier.
	// This should make things a bit quicker too!
	//

	private byte[,] pattern={
	{ 0, 32,  8, 40,  2, 34, 10, 42},   //* 8x8 Bayer ordered dithering  */
	{48, 16, 56, 24, 50, 18, 58, 26},   //* pattern.  Each input pixel   */
	{12, 44,  4, 36, 14, 46,  6, 38},   //* is scaled to the 0..63 range */
	{60, 28, 52, 20, 62, 30, 54, 22},   //* before looking in this table */
	{ 3, 35, 11, 43,  1, 33,  9, 41},   //* to determine the action.     */
	{51, 19, 59, 27, 49, 17, 57, 25},
	{15, 47,  7, 39, 13, 45,  5, 37},
	{63, 31, 55, 23, 61, 29, 53, 21} };

	public void ordered(byte[] gray, int w,  int h)
	{ 
		// byte col;
		int pos=0,prow=0,pcol=0;
		
		for(int y=0;y<h;y++)
		{
			for(int x=0;x<w;x++)
			{
				byte col=(byte)(gray[pos]>>2);
				if (col<pattern[pcol,prow]) 
				{
					gray[pos]=3;
					gray[pos]-=(byte)(col>>4); // 3- col>>4
				}
				else
				{
					gray[pos]=0;
				}

				pcol++;
				if (pcol>7){pcol=0;}
				
				pos++;	
			}

			prow++;
			if (prow>7){prow=0;}
		}
	}

	//
	// Error Diffusion
	// ===============
	// Also moved the Cybiko Colour array (cycol) out of the function to reduce allocations and make the GC happier.
	// Again, this should make things a little quicker too!
	//

	private byte[] cycol={255,171,85,0}; //0,85,171,255
		
	public void diffuse(byte[] gray, int w,  int h)
	{ 
		int pos=0,tc;
		byte closest,no;
		int carry;
		int best,dif;
		
		for(int y=0;y<h;y++)
		{
			carry=0;
			for(int x=0;x<w;x++)
			{
				best=999;
				closest=0;
				tc=gray[pos]+carry;
				for(no=0;no<4;no++)
				{
					dif=Math.Abs(cycol[no]-tc);
					if (dif<best)
					{
						best=dif;
						closest=no;
						carry=tc-cycol[no];
					}
				}
				gray[pos]=closest;
				pos++;
			}	
		}
	}

	//
	// Converts 24bit colour to 8bit gray (and flips)
	//

	public void grey(byte[] bmp,byte[]gray, int w,  int h)
	{
		int pos=0;
		
		for(int y=0;y<h;y++)
		{
			int c=((h-1)-y)*(w*3);	// flip image
		
			for(int x=0;x<w;x++)
			{
				byte r=bmp[c];
				byte g=bmp[c+1];
				byte b=bmp[c+2];
				gray[pos]=(byte)((r+g+b)/3);
			
				c+=3;
				pos++;
			}	
		}
	}



	public void twobit(byte[] gray, int w,  int h){
		for(int i=0;i<gray.Length;i++){
			gray[i]=(byte)(3-(gray[i]>>6));
		}
	}


	//
	// Converts frame from 8bit (gray) to 2bit
	//

	public void frameconv(byte[] gray,byte[] output, int w,  int h){
		int c=0,pos=0;
		
		for(int y=0;y<h;y++){
			for(int x=0;x<(w/4);x++){
				byte pa=gray[c];
				byte pb=gray[c+1];
				byte pc=gray[c+2];
				byte pd=gray[c+3];
				c=c+4;
				
				output[pos]=(byte)(pa<<6);
				output[pos]+=(byte)(pb<<4);
				output[pos]+=(byte)(pc<<2);
				output[pos]+=(byte)(pd);
				pos++;
			}	
		}
	}
}